@@ -41,6 +41,7 @@ module Text.LLVM (
     -- * Basic Blocks
   , BB()
   , runBB
+  , bbStmtModifier
   , freshLabel
   , label
   , comment
@@ -314,9 +315,36 @@ define' attrs rty sym sig va k = do
 -- Basic Block Monad -----------------------------------------------------------
 
 newtype BB a = BB
-  { unBB :: WriterT [BasicBlock] (StateT RW Id) a
+  { unBB :: ReaderT (Stmt -> Stmt) (WriterT [BasicBlock] (StateT RW Id)) a
   } deriving (Functor,Applicative,Monad,MonadFix)
 
+
+-- | The 'bbStmtModifier' function can be used to register a function that can
+-- modify the subsequent statements generated into this block.
+--
+-- For example, the following 'BB' monad code segment will emit a couple of LLVM
+-- statements:
+--
+-- > v <- load (iT 8) globalVar Nothing
+-- > call fooFunc [v]
+-- > jump end
+--
+-- But these statements will be \"plain\" in the resulting 'BasicBlock'.  If the
+-- caller wishes to add debug Metadata for location, they could instead write:
+--
+-- > bbStmtModifier (extendMetadata ("dbg", ValMdRef i))
+-- >  v <- load (iT 8) globalVar Nothing
+-- > bbStmtModifier (extendMetadata ("dbg", ValMdRef j))
+-- > call fooFunc [v]
+-- > jump end
+--
+-- Where @i@ and @j@ are the metadata index values of the 'DebugLoc' entries
+-- describing the source location of the \"load\" and \"call\"+\"jump\" statements,
+-- respectively.
+
+bbStmtModifier :: (Stmt -> Stmt) -> BB a -> BB a
+bbStmtModifier stmtModifier = BB . local stmtModifier . unBB
+
 avoidName :: String -> BB ()
 avoidName name = BB $ do
   rw <- get
@@ -333,7 +361,7 @@ freshNameBB pfx = BB $ do
 
 runBB :: BB a -> (a,[BasicBlock])
 runBB m =
-  case runId (runStateT emptyRW (runWriterT (unBB body))) of
+  case runId (runStateT emptyRW (runWriterT (runReaderT id (unBB body)))) of
     ((a,bbs),_rw) -> (a,bbs)
   where
   -- make sure that the last block is terminated
@@ -367,17 +395,18 @@ emitStmt :: Stmt -> BB ()
 emitStmt stmt = do
   BB $ do
     rw <- get
-    set $! rw { rwStmts = rwStmts rw Seq.|> stmt }
+    smod <- ask
+    set $! rw { rwStmts = rwStmts rw Seq.|> smod stmt }
   when (isTerminator (stmtInstr stmt)) terminateBasicBlock
 
 effect :: Instr -> BB ()
-effect i = emitStmt (Effect i [])
+effect i = emitStmt (Effect i mempty [])
 
 observe :: Type -> Instr -> BB (Typed Value)
 observe ty i = do
   name <- freshNameBB "r"
   let res = Ident name
-  emitStmt (Result res i [])
+  emitStmt (Result res i mempty [])
   return (Typed ty (ValIdent res))
 
 
@@ -508,8 +537,8 @@ assign r@(Ident name) body = do
   rw <- BB get
   case Seq.viewr (rwStmts rw) of
 
-    stmts Seq.:> Result _ i m ->
-      do BB (set rw { rwStmts = stmts Seq.|> Result r i m })
+    stmts Seq.:> Result _ i d m ->
+      do BB (set rw { rwStmts = stmts Seq.|> Result r i d m })
          return (const (ValIdent r) `fmap` tv)
 
     _ -> error "assign: invalid argument"
@@ -640,10 +669,10 @@ convop :: IsValue a
 convop k a ty = observe ty (k (toValue `fmap` a) ty)
 
 trunc :: IsValue a => Typed a -> Type -> BB (Typed Value)
-trunc  = convop (Conv Trunc)
+trunc  = convop (Conv (Trunc False False))
 
 zext :: IsValue a => Typed a -> Type -> BB (Typed Value)
-zext  = convop (Conv ZExt)
+zext  = convop (Conv (ZExt False))
 
 sext :: IsValue a => Typed a -> Type -> BB (Typed Value)
 sext  = convop (Conv SExt)
@@ -661,7 +690,7 @@ fptosi :: IsValue a => Typed a -> Type -> BB (Typed Value)
 fptosi  = convop (Conv FpToSi)
 
 uitofp :: IsValue a => Typed a -> Type -> BB (Typed Value)
-uitofp  = convop (Conv UiToFp)
+uitofp  = convop (Conv (UiToFp False))
 
 sitofp :: IsValue a => Typed a -> Type -> BB (Typed Value)
 sitofp  = convop (Conv SiToFp)
@@ -676,7 +705,7 @@ bitcast :: IsValue a => Typed a -> Type -> BB (Typed Value)
 bitcast  = convop (Conv BitCast)
 
 icmp :: (IsValue a, IsValue b) => ICmpOp -> Typed a -> b -> BB (Typed Value)
-icmp op l r = observe (iT 1) (ICmp op (toValue `fmap` l) (toValue r))
+icmp op l r = observe (iT 1) (ICmp False op (toValue `fmap` l) (toValue r))
 
 fcmp :: (IsValue a, IsValue b) => FCmpOp -> Typed a -> b -> BB (Typed Value)
 fcmp op l r = observe (iT 1) (FCmp op (toValue `fmap` l) (toValue r))
@@ -696,7 +725,7 @@ select c t f = observe (typedType t)
 
 getelementptr :: IsValue a
               => Type -> Typed a -> [Typed Value] -> BB (Typed Value)
-getelementptr ty ptr ixs = observe ty (GEP False ty (toValue `fmap` ptr) ixs)
+getelementptr ty ptr ixs = observe ty (GEP [] ty (toValue `fmap` ptr) ixs)
 
 -- | Emit a call instruction, and generate a new variable for its result.
 call :: IsValue a => Typed a -> [Typed Value] -> BB (Typed Value)
